# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import models

from parler.models import TranslatableModel, TranslatedFields
from cms.models.fields import PlaceholderField

from aldryn_reversion.core import version_controlled_content


@version_controlled_content
class SimpleRegistered(models.Model):
    """
    Represents the simplest possible model.
    """
    position = models.IntegerField()


@version_controlled_content
class SimpleNoAdmin(models.Model):
    """
    Represents the simplest possible model, but with no admin class, so
    that user cannot access revert view.
    """
    position = models.IntegerField()


@version_controlled_content
class WithTranslations(TranslatableModel):
    """
    Simple model with translated fields.
    """
    translations = TranslatedFields(
        description=models.CharField(max_length=255)
    )
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)


@version_controlled_content
class WithPlaceholder(models.Model):
    """
    Simple model with PlaceholderField.
    """
    content = PlaceholderField('WithPlaceholder Content ph')


@version_controlled_content(follow=['simple_relation'])
class SimpleFK(models.Model):
    """
    Simple relation to model which is not registered with reversions admin,
    and can be restored only in automatical mode.
    """
    simple_relation = models.ForeignKey(SimpleNoAdmin)


@version_controlled_content(follow=['simple_relation'])
class SimpleRequiredFK(models.Model):
    """
    Simple relation to model which is registered with reversions admin,
    and FK is required
    """
    simple_relation = models.ForeignKey(SimpleRegistered, null=True,
                                        on_delete=models.SET_NULL)


@version_controlled_content(follow=['simple_relation'])
class ComplexOneFK(TranslatableModel):
    """
    FK, placeholders, translations
    """
    translations = TranslatedFields(
        complex_description=models.TextField(max_length=500)
    )
    complex_content = PlaceholderField('ComplexOneFK complex_content ph')
    simple_relation = models.ForeignKey(WithPlaceholder)
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)


@version_controlled_content(follow=['simple_relation'])
class BlankFK(models.Model):
    """
    Simple model with not required FK relation.
    """
    simple_relation = models.ForeignKey(SimpleRegistered, blank=True, null=True)


@version_controlled_content(follow=['first_relation', 'second_relation'])
class MultiLevelFK(models.Model):
    """
    Few levels of foreign key relations, to test multilevel recovery.
    Level means there is a relation to an object which is also related to
    other object, i.e.:
        MultiLevelFK -> SimpleFK -> SimpleNoAdmin and
        MultiLevelFK -> ComplexOneFK -> WithPlaceholder
    """
    first_relation = models.ForeignKey(
        SimpleFK, blank=True, null=True, on_delete=models.SET_NULL)
    second_relation = models.ForeignKey(ComplexOneFK)


@version_controlled_content(follow=['self_relation'])
class FKtoSelf(models.Model):
    """
    Recursive relations model.
    """
    self_relation = models.ForeignKey(
        'self', on_delete=models.CASCADE, related_name='incoming_relations',
        blank=True)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from reversion.revisions import default_revision_manager
from cms.models import Placeholder

from aldryn_reversion.test_helpers.test_app.models import (
    SimpleNoAdmin, SimpleFK, WithPlaceholder, WithTranslations,
)

from ..forms import RecoverObjectWithTranslationForm
from ..utils import (
    RecursiveRevisionConflictResolver, get_deleted_placeholders_for_object,
)

from .base import (
    ReversionBaseTestCase, HelperModelsObjectsSetupMixin,
    get_version_for_object
)


class FormsTestCase(HelperModelsObjectsSetupMixin, ReversionBaseTestCase):
    form_class = RecoverObjectWithTranslationForm

    def build_form_kwargs(self, obj):
        version = get_version_for_object(obj)
        return {
            'revision': version.revision,
            'obj': obj,
            'version': version,
            'resolve_conflicts': [],
            'placeholders': [],
        }

    def build_unbound_form(self, args_dict):
        """
        Build a self.form_class unbound form, args_dict - keyword arguments
        that will be passed to form class __init__.
        """
        return self.form_class(**args_dict)

    def build_bound_form(self, args_dict):
        """
        Build a self.form_class bound form by also providing `data` kw argument,
        to form_class __init__ method to bypass django checks.
        Expects args_dict to be keyword arguments that will be passed to
        form class __init__.
        """
        dummy = {'dummy': 1}
        return self.form_class(data=dummy, **args_dict)

    def compare_form_attrs(self, form, form_kwargs):
        """
        Check that form attributes equal to form_kwargs.
        tries to get attributes based on form_kwargs keys and compare to
        form_kwargs values. if none or miss match - returns dict of
        {attribute name: attribute value}
        """
        not_equal = {}
        for arg_name, value in form_kwargs.items():
            found = False
            form_arg = getattr(form, arg_name, None)
            if type(form_arg) == list:
                arg_not_equal = [item for item in form_arg
                                 if item not in form_kwargs[arg_name]]
                # if there is less or more items consider lists not equal
                if arg_not_equal or len(form_arg) > len(form_kwargs[arg_name]):
                    found = True
            elif form_arg is None or form_arg != form_kwargs[arg_name]:
                arg_not_equal = form_arg
                found = True
            if found:
                not_equal[arg_name] = arg_not_equal
        return not_equal

    def test_recover_form_init(self):
        # test with not translatable object, should not contain translations
        # field
        form_simple_kwargs = self.build_form_kwargs(self.simple_no_admin)
        form_simple = self.build_unbound_form(form_simple_kwargs)
        self.assertNotIn('translations', form_simple.fields.keys())

        # test with translatable object, should contain translations field
        form_translated_kwargs = self.build_form_kwargs(self.with_translation)
        form_translated = self.build_unbound_form(form_translated_kwargs)
        self.assertIn('translations', form_translated.fields.keys())

        # test that form has correct in memory attrs which are (atm):
        # * revision,
        # * obj (object which would be recovered)
        # * version (object version)
        # * resolve_conflicts (list of versions to recover)
        # * placeholders (list of versions to recover)
        # Also test that creating second form doesn't affect first form
        # in-memory attributes and vice versa.
        not_equal = self.compare_form_attrs(form_simple, form_simple_kwargs)
        message = 'Args not found or missmatch for {0} object {1}'
        self.assertEqual(len(not_equal), 0, msg=message.format(
            'simple_no_admin', not_equal))
        not_equal = self.compare_form_attrs(form_translated,
                                            form_translated_kwargs)
        self.assertEqual(len(not_equal), 0, msg=message.format(
            'with_translation', not_equal))

    def test_recover_form_clean(self):
        # test with no conflicts, should validate fine
        form_simple_kwargs = self.build_form_kwargs(self.simple_no_admin)
        # form validation doesn't triggers if form is not bound, so
        # we need to pass dummy data so it would be considered as bound
        # Should not be the case in real life because of request.POST

        form_simple = self.build_bound_form(form_simple_kwargs)
        self.assertTrue(form_simple.is_valid())

        # test with not resolved conflicts - should rise validation error
        simple_no_admin_version = default_revision_manager.get_for_object(
            self.simple_no_admin)[0]
        self.simple_no_admin.delete()
        self.assertEqual(SimpleNoAdmin.objects.count(), 0)
        self.assertEqual(SimpleFK.objects.count(), 0)
        simple_fk_kwargs = self.build_form_kwargs(self.simple_fk)
        form_simple_fk_with_conflicts = self.build_bound_form(simple_fk_kwargs)
        self.assertFalse(form_simple_fk_with_conflicts.is_valid())
        self.assertEqual(
            len(form_simple_fk_with_conflicts.non_field_errors()), 1)
        self.assertEqual(form_simple_fk_with_conflicts.non_field_errors()[0],
                         'Cannot restore object, there are conflicts!')

        # test with conflicts that were resolved by utility
        # RecursiveRevisionConflictResolver, and result were passed to form
        # init, should validate fine
        simple_fk_resolved_conflicts = RecursiveRevisionConflictResolver(
            simple_fk_kwargs['version']).resolve()
        simple_fk_kwargs['resolve_conflicts'] = simple_fk_resolved_conflicts
        form_simple_fk_resolved = self.build_bound_form(simple_fk_kwargs)
        self.assertTrue(form_simple_fk_resolved.is_valid())

        # test with conflicts that were resolved, and passed to form init,
        # but then something more was corrupted (new conflicts) should raise
        simple_no_admin_version.revert()
        partially_resolved = RecursiveRevisionConflictResolver(
            simple_fk_kwargs['version']).resolve()
        simple_fk_kwargs['resolve_conflicts'] = partially_resolved
        form_simple_fk_resolved = self.build_bound_form(simple_fk_kwargs)
        # delete object which was not resolved by resolver
        SimpleNoAdmin.objects.all()[0].delete()
        self.assertEqual(SimpleNoAdmin.objects.count(), 0)
        self.assertFalse(form_simple_fk_resolved.is_valid())

    def test_recover_form_save(self):
        # test reverts object_version (no extras, simple object with
        # no translations, placeholders or fks
        form_simple_kwargs = self.build_form_kwargs(self.simple_no_admin)
        self.simple_no_admin.delete()
        self.assertEqual(SimpleNoAdmin.objects.count(), 0)
        form_simple = self.build_bound_form(form_simple_kwargs)
        self.assertTrue(form_simple.is_valid())
        form_simple.save()
        self.assertEqual(SimpleNoAdmin.objects.count(), 1)

        # test reverts object and deleted placeholders
        with_placeholder_version = default_revision_manager.get_for_object(
            self.with_placeholder)[0]
        placeholder_pk = self.with_placeholder.content.pk
        self.with_placeholder.content.delete()
        # check that placeholder was actually deleted.
        self.assertEqual(
            Placeholder.objects.filter(pk=placeholder_pk).count(), 0)

        placeholder_versions = get_deleted_placeholders_for_object(
            self.with_placeholder, with_placeholder_version.revision)
        form_with_placeholders_kwargs = self.build_form_kwargs(
            self.with_placeholder)
        self.with_placeholder.delete()
        self.assertEqual(WithPlaceholder.objects.count(), 0)

        form_with_placeholders_kwargs['placeholders'] = placeholder_versions
        form_with_placeholder = self.build_bound_form(
            form_with_placeholders_kwargs)
        self.assertTrue(form_with_placeholder.is_valid())
        form_with_placeholder.save()
        self.assertEqual(WithPlaceholder.objects.count(), 1)
        # check that placeholder was restored
        self.assertEqual(
            Placeholder.objects.filter(pk=placeholder_pk).count(), 1)

        # test reverts object and translations
        form_translated_kwargs = self.build_form_kwargs(self.with_translation)

        self.with_translation.delete()
        self.assertEqual(WithTranslations.objects.count(), 0)
        self.assertEqual(
            self.with_translation.translations.count(), 0)
        form_with_translation = self.build_unbound_form(form_translated_kwargs)

        translations_pks = [tr[0] for tr in
                            form_with_translation.fields[
                                'translations'].choices]
        form_with_translation = self.form_class(
            {'translations': translations_pks}, **form_translated_kwargs)
        self.assertTrue(form_with_translation.is_valid())
        form_with_translation.save()

        self.assertEqual(WithTranslations.objects.count(), 1)
        self.assertEqual(WithTranslations.objects.get().translations.count(), 2)
        self.assertEqual(self.with_translation.description,
                         WithTranslations.objects.get().description)

        # test reverts object and conflicts
        simple_fk_kwargs = self.build_form_kwargs(self.simple_fk)
        SimpleNoAdmin.objects.get().delete()
        self.assertEqual(SimpleNoAdmin.objects.count(), 0)
        self.assertEqual(SimpleFK.objects.count(), 0)
        simple_fk_resolved_conflicts = RecursiveRevisionConflictResolver(
            simple_fk_kwargs['version']).resolve()
        simple_fk_kwargs['resolve_conflicts'] = simple_fk_resolved_conflicts
        form_simple_fk_resolved = self.build_bound_form(simple_fk_kwargs)
        self.assertTrue(form_simple_fk_resolved.is_valid())
        form_simple_fk_resolved.save()
        self.assertEqual(SimpleFK.objects.count(), 1)
        self.assertEqual(SimpleNoAdmin.objects.count(), 1)
        self.assertEqual(self.simple_fk.simple_relation,
                         SimpleFK.objects.get().simple_relation)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.contrib import admin

from parler.admin import TranslatableAdmin
from aldryn_reversion.admin import VersionedPlaceholderAdminMixin
from cms.admin.placeholderadmin import FrontendEditableAdminMixin

from .models import (
    SimpleRegistered, WithTranslations, WithPlaceholder, SimpleFK,
    ComplexOneFK, BlankFK, MultiLevelFK, FKtoSelf, SimpleRequiredFK
)


class SimpleRegisteredAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


class WithTranslationsAdmin(VersionedPlaceholderAdminMixin,
                            TranslatableAdmin):
    pass


class WithPlaceholdersAdmin(VersionedPlaceholderAdminMixin,
                            FrontendEditableAdminMixin,
                            admin.ModelAdmin):
    pass


class SimpleFKAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


class SimpleRequiredFKAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


class ComplexOneFKAdmin(VersionedPlaceholderAdminMixin,
                        FrontendEditableAdminMixin,
                        TranslatableAdmin):
    pass


class BlankFKAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


class MultiLevelFKAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


class FKtoSelfAdmin(VersionedPlaceholderAdminMixin, admin.ModelAdmin):
    pass


admin.site.register(SimpleRegistered, SimpleRegisteredAdmin)
admin.site.register(WithTranslations, WithTranslationsAdmin)
admin.site.register(WithPlaceholder, WithPlaceholdersAdmin)
admin.site.register(SimpleFK, SimpleFKAdmin)
admin.site.register(SimpleRequiredFK, SimpleRequiredFKAdmin)
admin.site.register(ComplexOneFK, ComplexOneFKAdmin)
admin.site.register(BlankFK, BlankFKAdmin)
admin.site.register(MultiLevelFK, MultiLevelFKAdmin)
admin.site.register(FKtoSelf, FKtoSelfAdmin)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import six

from reversion.models import Version
from reversion.revisions import (
    revision_context_manager, default_revision_manager)

from django.db import transaction
from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.contrib.auth.hashers import make_password
from django.contrib.messages.storage.fallback import FallbackStorage
from django.utils.translation import override
from django.test import RequestFactory, TransactionTestCase

from cms.utils.i18n import get_language_list

from aldryn_reversion.test_helpers.test_app.models import (
    SimpleNoAdmin, SimpleRegistered, WithTranslations, WithPlaceholder,
    SimpleFK, BlankFK, ComplexOneFK, MultiLevelFK, SimpleRequiredFK,
)


def get_version_for_object(obj, revision_pk=None):
    """
    Return version for object, if revision_pk is provided - returns a version
    from that specific revision. No exception handling! if you get exceptions
    check your logic.
    :param obj: revisionable object
    :param revision_pk: int revision pk
    :return: reversion.models.Version
    """
    versions = default_revision_manager.get_for_object(obj)
    if revision_pk is not None:
        versions = versions.filter(revision__pk=revision_pk)
    version = versions[0]
    return version


class ReversionBaseTestCase(TransactionTestCase):

    raw_text_string = {
        'en': 'just {0} raw string',
        'de': 'just DE {0} raw string',
    }

    def setUp(self):
        super(ReversionBaseTestCase, self).setUp()

        self.staff_user_password = 'staff_pw'
        self.staff_user = self.create_staff_user(
            'staff', self.staff_user_password)
        self.super_user_password = 'super_pw'
        self.super_user = self.create_super_user(
            'super', self.super_user_password)

    def create_user(self, user_name, user_password, is_staff=False,
                    is_superuser=False):
        return User.objects.create(
            username=user_name,
            first_name='{0} first_name'.format(user_name),
            last_name='{0} last_name'.format(user_name),
            password=make_password(user_password),
            is_staff=is_staff,
            is_superuser=is_superuser
        )

    def create_staff_user(self, user_name, user_password):
        staff_user = self.create_user(user_name, user_password, is_staff=True)
        return staff_user

    def create_super_user(self, user_name, user_password):
        super_user = self.create_user(user_name, user_password,
                                      is_superuser=True)
        return super_user

    def create_with_revision(self, kls, language='en', **kwargs):
        """
        Create an instance of class kls with passing kwargs to
        kls.objects.create(**kwargs)
        """
        with transaction.atomic():
            with revision_context_manager.create_revision():
                with override(language):
                    return kls.objects.create(**kwargs)

    def create_translation_with_revision(self, language, obj, **kwargs):
        """
        Just creates a translation under revision manager. Returns nothing.
        """
        # create translation outside of revision manager, and save the object
        # after that, so that revision would contain full set of related
        # objects, instead of containing only translation and no master object.
        obj.create_translation(language, **kwargs)
        with transaction.atomic():
            with revision_context_manager.create_revision():
                obj.save()

    def create_revision(self, obj, content=None, **kwargs):
        with transaction.atomic():
            with revision_context_manager.create_revision():
                # populate event with new values
                for property, value in six.iteritems(kwargs):
                    setattr(obj, property, value)
                if content:
                    # get correct plugin for language. do not update the same
                    # one.
                    language = obj.get_current_language()
                    plugins = obj.content.get_plugins().filter(
                        language=language)
                    plugin = plugins[0].get_plugin_instance()[0]
                    plugin.body = content
                    plugin.save()
                obj.save()

    def revert_to(self, object_with_revision, revision_number):
        """
        Revert <object with revision> to revision number.
        """
        # get by position, since reversion_id is not reliable,
        version = list(reversed(
            default_revision_manager.get_for_object(
                object_with_revision)))[revision_number - 1]
        version.revision.revert()

    def get_object(self, object_or_version):
        """
        Returns object regardless of type fo object_or_version.
        """
        return (object_or_version
                if type(object_or_version) != Version
                else object_or_version.object_version.object)

    def get_version(self, object_or_version):
        """
        Returns version or latest version for object, if object_or_version is
        object not reversion.models.Version.
        """
        return (object_or_version if type(object_or_version) == Version
                else get_version_for_object(object_or_version))


class HelperModelsObjectsSetupMixin(object):

    def setUp(self):
        super(HelperModelsObjectsSetupMixin, self).setUp()
        # prepare objects that can be used with utility methods
        # simple model
        self.simple_registered = self.create_with_revision(
            SimpleRegistered, position=1)

        self.simple_no_admin = self.create_with_revision(
            SimpleNoAdmin, position=42)

        # translations
        # note that WithTranslations will have 2 revisions, initial one
        # and with 'de' translation
        en_description = self.raw_text_string['en'].format(0)
        de_description = self.raw_text_string['de'].format(0)
        self.with_translation = self.create_with_revision(
            WithTranslations, description=en_description)
        self.create_translation_with_revision(
            'de', self.with_translation, description=de_description)

        # placeholders
        self.with_placeholder = self.create_with_revision(
            WithPlaceholder)

        # FK
        self.simple_fk = self.create_with_revision(
            SimpleFK, simple_relation=self.simple_no_admin)
        # blank FK means that fk is not required.
        self.blank_fk = self.create_with_revision(
            BlankFK,
            simple_relation=self.simple_registered)
        self.simple_required_fk = self.create_with_revision(
            SimpleRequiredFK, simple_relation=self.simple_registered)
        # ComplexOneFK
        self.complex_one_fk = self.create_with_revision(
            ComplexOneFK,
            simple_relation=self.with_placeholder,
            complex_description=en_description)
        self.create_translation_with_revision(
            'de', self.complex_one_fk, complex_description=de_description)
        # MultiLevelFK
        self.multi_level_fk = self.create_with_revision(
            MultiLevelFK,
            first_relation=self.simple_fk,
            second_relation=self.complex_one_fk)


class CMSRequestBasedMixin(object):
    languages = get_language_list()

    @classmethod
    def setUpClass(cls):
        super(CMSRequestBasedMixin, cls).setUpClass()
        cls.request_factory = RequestFactory()
        cls.site1 = Site.objects.get(pk=1)

    @staticmethod
    def get_request(language=None, url="/"):
        """
        Returns a Request instance populated with cms specific attributes.
        User is not set.
        """
        request_factory = RequestFactory(HTTP_HOST=settings.ALLOWED_HOSTS[0])
        request = request_factory.get(url)
        request.LANGUAGE_CODE = language or settings.LANGUAGE_CODE
        # Needed for plugin rendering.
        request.current_page = None
        # session and messages
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        setattr(request, '_messages', messages)
        return request

    def get_su_request(self, language=None, url="/"):
        request = self.get_request(language, url)
        request.user = self.super_user
        return request
